﻿
using System.ComponentModel;
using System.Linq;
using HeuristicLab.Algorithms.DataAnalysis;
using HeuristicLab.Algorithms.OffspringSelectionGeneticAlgorithm;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Encodings.ParameterConfigurationTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
namespace HeuristicLab.ExperimentGeneration.DataAnalysis.ExperimentWizard {
  public partial class MediumAnalysisPage : HeuristicLab.ExperimentGeneration.DataAnalysis.Wizard.WizardPage {
    private const int nrOfRepetitions = 10;
    private const string gaCrossover = "SubtreeSwappingCrossover";
    private const string gaMutator = "MultiSymbolicExpressionTreeManipulator";
    private const string gaSelector = "GenderSpecificSelection";

    private BackgroundWorker worker;

    private DataAnalysisWizardContext context;
    public DataAnalysisWizardContext Context {
      get { return context; }
    }

    public MediumAnalysisPage(DataAnalysisWizardContext context) {
      InitializeComponent();
      this.context = context;
      worker = new BackgroundWorker();
      worker.DoWork += new DoWorkEventHandler(GenerateExperimentsTask);
      worker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(WorkerCompleted);
      worker.WorkerReportsProgress = true;
      worker.ProgressChanged += new ProgressChangedEventHandler(WorkerProgressChanged);

    }

    private void GenerateExperimentsTask(object sender, DoWorkEventArgs e) {
      DataAnalysisWizardContext c = e.Argument as DataAnalysisWizardContext;
      IProblem problem = c.Problem;

      // SVR Experiment -------------------------------------------------------
      //SupportVectorRegression svr = new SupportVectorRegression();
      //svr.Problem = (IRegressionProblem)problem;
      //ParameterConfigurationTree vc = new ParameterConfigurationTree(svr, problem);

      //var nuRange = new DoubleValueRange(new DoubleValue(0.1), new DoubleValue(0.9), new DoubleValue(0.1));
      //SetParameterRangeContraint(vc, "Nu", nuRange);

      //var costRange = new DoubleValueFactorRange(new DoubleValue(0.03125), new DoubleValue(32768), new DoubleValue(4));
      //SetParameterRangeContraint(vc, "Cost", costRange);

      //var gammaRange = new DoubleValueFactorRange(new DoubleValue(6.10352E-05), new DoubleValue(64), new DoubleValue(4));
      //SetParameterRangeContraint(vc, "Gamma", gammaRange);
      //worker.ReportProgress(5);
      //Experiment svrExperiment = vc.GenerateExperiment(svr);
      // ======================================================================
      worker.ReportProgress(20);

      // RF Experiment --------------------------------------------------------
      //RandomForestRegression rfr = new RandomForestRegression();
      //rfr.Problem = (IRegressionProblem)problem;
      //ParameterConfigurationTree rfConfig = new ParameterConfigurationTree(rfr, problem);
      //var rRange = new DoubleValueRange(new DoubleValue(0.1), new DoubleValue(0.7), new DoubleValue(0.1));
      //SetParameterRangeContraint(rfConfig, "R", rRange);
      //var treeRange = new IntValueRange(new IntValue(250), new IntValue(250), new IntValue(250));
      //SetParameterRangeContraint(rfConfig, "Number of trees", treeRange);
      //Experiment rfrExperiment = rfConfig.GenerateExperiment(rfr);
      // ======================================================================
      worker.ReportProgress(40);
      // NN Experiment --------------------------------------------------------
      //NeuralNetworkRegression nnr = new NeuralNetworkRegression();
      //nnr.Problem = (IRegressionProblem)problem;
      //ParameterConfigurationTree nnrConfig = new ParameterConfigurationTree(nnr, problem);
      //var decayRange = new DoubleValueFactorRange(new DoubleValue(0.001), new DoubleValue(100), new DoubleValue(10));
      //SetParameterRangeContraint(nnrConfig, "Decay", decayRange);
      //var nodesRange = new IntValueFactorRange(new IntValue(1), new IntValue(100), new IntValue(3));
      //SetParameterRangeContraint(nnrConfig, "NodesInFirstHiddenLayer", nodesRange);
      //Experiment nnrExperiment = nnrConfig.GenerateExperiment(nnr);
      // ======================================================================
      worker.ReportProgress(60);
      // GP Experiment --------------------------------------------------------
      OffspringSelectionGeneticAlgorithm osga = new OffspringSelectionGeneticAlgorithm();
      var prob = new SymbolicRegressionSingleObjectiveProblem();
      prob.ProblemData = ((IRegressionProblem)problem).ProblemData;
      //prob.SolutionCreator = new MultiSymbolicDataAnalysisExpressionCreator();
      osga.Problem = prob;
      osga.ComparisonFactorLowerBound.Value = 1;
      osga.Crossover = osga.Problem.Operators.OfType<ICrossover>().FirstOrDefault(x => x.Name == gaCrossover);
      osga.MaximumEvaluatedSolutions.Value = 500000;
      osga.MaximumGenerations.Value = 100;
      osga.MutationProbability = new PercentValue(0.15);
      osga.Mutator = osga.Problem.Operators.OfType<IManipulator>().FirstOrDefault(x => x.Name == gaMutator);
      osga.PopulationSize.Value = 500;
      osga.Selector = osga.SelectorParameter.ValidValues.FirstOrDefault(x => x.Name == gaSelector);

      CrossValidation crossSmall = new CrossValidation();
      crossSmall.Algorithm = osga;
      BatchRun batchSmall = new BatchRun("small");
      batchSmall.Repetitions = nrOfRepetitions;
      batchSmall.Optimizer = crossSmall;

      CrossValidation crossMedium = (CrossValidation)crossSmall.Clone(new Cloner());
      var probMedium = (SymbolicRegressionSingleObjectiveProblem)prob.Clone(new Cloner());
      probMedium.MaximumSymbolicExpressionTreeDepth.Value = 12;
      probMedium.MaximumSymbolicExpressionTreeLength.Value = 125;
      crossMedium.Algorithm.Problem = probMedium;
      BatchRun batchMedium = new BatchRun("medium");
      batchMedium.Repetitions = nrOfRepetitions;
      batchMedium.Optimizer = crossMedium;

      CrossValidation crossLarge = (CrossValidation)crossSmall.Clone(new Cloner());
      var probLarge = (SymbolicRegressionSingleObjectiveProblem)prob.Clone(new Cloner());
      probLarge.MaximumSymbolicExpressionTreeDepth.Value = 17;
      probLarge.MaximumSymbolicExpressionTreeLength.Value = 250;
      crossLarge.Algorithm.Problem = probLarge;
      BatchRun batchLarge = new BatchRun("large");
      batchLarge.Repetitions = nrOfRepetitions;
      batchLarge.Optimizer = crossLarge;

      Experiment gpExperiment = new Experiment();
      gpExperiment.Optimizers.Add(batchSmall);
      gpExperiment.Optimizers.Add(batchMedium);
      gpExperiment.Optimizers.Add(batchLarge);
      // ======================================================================
      worker.ReportProgress(80);

      Experiment experiment = new Experiment("Data Analysis Experiment");
      //experiment.Optimizers.Add(svrExperiment);
      //experiment.Optimizers.Add(rfrExperiment);
      //experiment.Optimizers.Add(nnrExperiment);
      experiment.Optimizers.Add(gpExperiment);

      worker.ReportProgress(100);

      e.Result = experiment;
    }

    private void SetParameterRangeContraint(ParameterConfigurationTree pct, string parameterName, IRange rangeContraint) {
      var pc = pct.AlgorithmConfiguration.ParameterConfigurations.Where(x => x.ParameterName == parameterName).SingleOrDefault();
      pc.Optimize = true;
      var vc = (RangeValueConfiguration)pc.ValueConfigurations.First();
      vc.Optimize = true;
      vc.RangeConstraint = rangeContraint;

    }

    private void WorkerProgressChanged(object sender, ProgressChangedEventArgs e) {
      progressBar1.Value = e.ProgressPercentage;
    }

    private void WorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
      SetWizardButton(Wizard.WizardButtons.Next);
      Experiment experiment = (Experiment)e.Result;
      context.Experiment = experiment;
    }

    private void MediumAnalysisPage_WizardBack(object sender, Wizard.WizardPageEventArgs e) {
      e.NewPage = "SelectAnalysisPage";
    }

    private void MediumAnalysisPage_SetActive(object sender, System.ComponentModel.CancelEventArgs e) {
      SetWizardButton(Wizard.WizardButtons.None);
      worker.RunWorkerAsync(context);
    }
  }
}
